# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from typing import List, Optional, Tuple

import matplotlib.pyplot as plt
from torch import Tensor


def get_axis(axarr, H, W, i, j):
    H, W = H - 1, W - 1
    if not (H or W):
        ax = axarr
    elif not (H and W):
        ax = axarr[max(i, j)]
    else:
        ax = axarr[i][j]
    return ax


def show_image_row(
    xlist: Tensor,
    ylist: Optional[List[str]] = None,
    fontsize: int = 12,
    size: Tuple[float, float] = (2.5, 2.5),
    tlist: Optional[List[str]] = None,
    filename: Optional[str] = None,
):
    H, W = len(xlist), len(xlist[0])
    fig, axarr = plt.subplots(
        H, W, figsize=(size[0] * W, size[1] * H), subplot_kw=dict(xticks=[], yticks=[])
    )
    for w in range(W):
        for h in range(H):
            ax = get_axis(axarr, H, W, h, w)
            im = xlist[h][w]
            if len(im) == 1:
                im = im.repeat(3, 1, 1)

            ax.imshow(im.permute(1, 2, 0))
            if ylist and w == 0:
                ax.set_ylabel(ylist[h], fontsize=fontsize)
            if tlist:
                ax.set_title(tlist[h][w], fontsize=fontsize)
    if filename is not None:
        plt.savefig(filename, bbox_inches="tight")
    return fig



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from typing import Any

import torch
from torch import Tensor
from torchvision.models import resnet

from ._resnet_blocks import VizBasicBlock, VizBottleneck


class ResNet(resnet.ResNet):
    def __init__(self, *args, with_latent=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.with_latent = with_latent

    def _forward_impl(self, x: Tensor) -> Tensor:
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        if self.with_latent:
            return x

        x = self.fc(x)
        return x

    def forward(self, x: Tensor) -> Tensor:
        return self._forward_impl(x)


def resnet18(viz_relu: bool = False, **kwargs: Any) -> ResNet:
    """ResNet-18 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_.
    """
    block = resnet.BasicBlock
    if viz_relu:
        block = VizBasicBlock
    return ResNet(block, [2, 2, 2, 2], **kwargs)


def resnet50(viz_relu: bool = False, **kwargs: Any) -> ResNet:
    """ResNet-50 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_.
    """
    block = resnet.Bottleneck
    if viz_relu:
        block = VizBottleneck
    return ResNet(block, [3, 4, 6, 3], **kwargs)



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from typing import Any, Callable, List, Optional, Type, Union

import torch
from torch import Tensor, nn
from torchvision.models import resnet
from torchvision.models.resnet import BasicBlock, Bottleneck

from ._resnet_blocks import VizBasicBlock, VizBottleneck


class ResNet(resnet.ResNet):
    """TorchVision ResNet Adapted for smaller images (i.e., CIFAR-10).

    Parameters
    ----------
    block: Type[BasicBlock] | Type[Bottleneck]

    layers: List[int]

    num_classes: int (default: 10)

    zero_init_residual: bool (default: False)

    groups: int (default: 1)

    width_per_group: int (default: 64)

    replace_stride_with_dilation: List[bool] | None (default: None)

    norm_layer: Callable[..., nn.Module] | None (default: None)

    input_channels: int (default: 3)
    """

    def __init__(
        self,
        block: Type[Union[BasicBlock, Bottleneck]],
        layers: List[int],
        num_classes: int = 10,
        zero_init_residual: bool = False,
        groups: int = 1,
        width_per_group: int = 64,
        replace_stride_with_dilation: Optional[List[bool]] = None,
        norm_layer: Optional[Callable[..., nn.Module]] = None,
        input_channels: int = 3,
    ):
        super().__init__(
            block,
            layers,
            num_classes,
            zero_init_residual,
            groups,
            width_per_group,
            replace_stride_with_dilation,
            norm_layer,
        )
        del self.maxpool

        self.conv1 = nn.Conv2d(
            input_channels,
            64,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=False,
        )

        self.fc = nn.Linear(self.inplanes, 10)

    def _forward_impl(self, x: Tensor) -> Tensor:
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.fc(x)

        return x


def resnet18(viz_relu: bool = False, **kwargs: Any) -> ResNet:
    """ResNet-18 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_.
    """
    block = BasicBlock
    if viz_relu:
        block = VizBasicBlock
    return ResNet(block, [2, 2, 2, 2], **kwargs)


def resnet34(viz_relu: bool = False, **kwargs: Any) -> ResNet:
    """ResNet-34 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_.
    """
    block = BasicBlock
    if viz_relu:
        block = VizBasicBlock
    return ResNet(block, [3, 4, 6, 3], **kwargs)


def resnet50(viz_relu: bool = False, **kwargs: Any) -> ResNet:
    """ResNet-50 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_.
    """
    block = Bottleneck
    if viz_relu:
        block = VizBottleneck
    return ResNet(block, [3, 4, 6, 3], **kwargs)


def wide_resnet18_2(viz_relu: bool = False, **kwargs: Any) -> ResNet:
    """Wide ResNet-18 model"""
    kwargs["width_per_group"] = 64 * 2
    block = BasicBlock
    if viz_relu:
        block = VizBasicBlock
    return ResNet(block, [2, 2, 2, 2], **kwargs)


def wide_resnet50_2(viz_relu: bool = False, **kwargs: Any) -> ResNet:
    """Wide ResNet-50 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_.
    """
    kwargs["width_per_group"] = 64 * 2
    block = Bottleneck
    if viz_relu:
        block = VizBottleneck
    return ResNet(block, [3, 4, 6, 3], **kwargs)



from functools import partial

import pooch
from typing_extensions import Literal

__all__ = ["load_model"]


_pre_trained_manager = pooch.create(
    path=pooch.os_cache("rai-experiments"),
    base_url="https://github.com/mit-ll-responsible-ai/data_hosting/releases/download/madrylab_models/",
    registry={
        "mitll_cifar_l2_1_0.pt": "md5:738f109685dfc55f7227dc6dc473379e",
        "mitll_cifar_nat.pt": "md5:34e15ff183735933e13f41ad27f755b4",
        "mitll_restricted_imagenet_l2_3_0.pt": "md5:db37145acfd8bdf7ef7c05f4d261c0d9",
        "mitll_imagenet_l2_3_0.pt": "md5:37728e5c9c47684c0edd373e9d80ac9b",
    },
)


def load_model(
    model_name: Literal[
        "mitll_cifar_l2_1_0.pt",
        "mitll_cifar_nat.pt",
        "mitll_imagenet_l2_3_0.pt",
        "mitll_restricted_imagenet_l2_3_0.pt",
        "imagenet_nat.pt",
    ]
):
    r"""
    Loads pre-trained model weights. This function takes care of downloading and caching
    weights.

    All model weights were provided by Madry Lab [1]_. We converted these weights to a
    format that does not require extraneous dependencies, such as `dill`, to load.

    Parameters
    ----------
    model_name: str
        Supported models
           - mitll_cifar_l2_1_0.pt
           - mitll_cifar_nat.pt
           - mitll_imagenet_l2_3_0.pt
           - mitll_restricted_imagenet_l2_3_0.pt
           - imagenet_nat.p

    Returns
    -------
    loaded_model : torch.nn.Module
        The loaded model. (Note that the model is **not** placed in eval-mode by default).

    Notes
    -----
    Descriptions of models:

    - `mitll_cifar_nat.pt`: This is a ResNet-50 model trained on CIFAR 10 using standard training with no adversarial perturbations in the loop (i.e., :math:`\epsilon=0`)
    - `mitll_cifar_l2_1_0.pt`: A ResNet-50 model trained on CIFAR 10 with perturbations generated via PGD using perturbations constrained to :math:`L^2`-ball of radius :math:`\epsilon=1.0`
    -  `mitll_imagenet_l2_3_0`: This is a ResNet-50 model trained on ImageNet with PGD using :math:`\epsilon=3.0`
    -  `mitll_restricted_imagenet_l2_3_0`: This is a ResNet-50 model trained on restricted ImageNet [2]_ with PGD using :math:`\epsilon=3.0`

    References
    ----------
    .. [1] https://github.com/MadryLab/robustness
    .. [2] https://github.com/MadryLab/robust_representations
    """

    import torch
    from torchvision import transforms
    from torchvision.models import resnet50 as torchvision_resnet50

    from rai_toolbox.mushin._utils import load_from_checkpoint

    from ..models.resnet import resnet50
    from ..models.small_resnet import resnet50 as small_resnet50

    if model_name in {"mitll_cifar_l2_1_0.pt", "mitll_cifar_nat.pt"}:
        model = small_resnet50
        norm = transforms.Normalize(
            mean=[0.4914, 0.4822, 0.4465],
            std=[0.2023, 0.1994, 0.2010],
        )
    elif model_name in {
        "mitll_imagenet_l2_3_0.pt",
        "mitll_restricted_imagenet_l2_3_0.pt",
        "imagenet_nat.pt",
    }:

        model = partial(resnet50, num_classes=9 if "restricted" in model_name else 1000)
        norm = transforms.Normalize(
            mean=[0.485, 0.456, 0.406],
            std=[0.229, 0.224, 0.225],
        )
    else:
        raise ValueError(
            f"Unknown model name: {model_name}\nAvailable models: {', '.join(_pre_trained_manager.registry_files + ['imagenet_nat.pt'])}"
        )

    if model_name == "imagenet_nat.pt":
        base_model = torchvision_resnet50(pretrained=True)
    else:
        base_model = load_from_checkpoint(
            model=model(),
            ckpt=_pre_trained_manager.fetch(model_name),
            weights_key="state_dict",
        )

    model = torch.nn.Sequential(norm, base_model)
    return model



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

import torch
from torch import Tensor
from torchvision.models.resnet import BasicBlock, Bottleneck


class VizReLU(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input):
        return input.clamp(min=0)

    @staticmethod
    def backward(ctx, grad_output):
        return grad_output


class VizBasicBlock(BasicBlock):
    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = VizReLU.apply(out)

        return out


class VizBottleneck(Bottleneck):
    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = VizReLU.apply(out)

        return out



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from setuptools import find_packages, setup

INSTALL_REQUIRES = [
    #"matplotlib >= 2.0.0",
    #"hydra_zen >= 0.7.0",
    #"pytorch-lightning >= 1.5.0",
    #"protobuf <= 3.20.1",
    "pooch >= 1.6.0",
]


DESCRIPTION = (
    "Convenience functions and utilities for rai-toolbox tutorials and demonstrations"
)

LONG_DESCRIPTION = """
rai-experiments provides convenience functions and utilities for rai-toolbox
tutorials and demonstrations.

Note that the contents of rai-experiments are not tested or maintained to
the same fidelity as rai-toolbox. Thus this package is not meant to be relied
on in production. Use at your own risk!

**Installation**

```console
$ pip install rai-experiments"
```
"""

AUTHOR = "Olivia Brown, Justin Goodwin, Ryan Soklaski, Michael Yee"
AUTHOR_EMAIL = "ryan.soklaski@ll.mit.edu"

setup(
    name="rai_experiments",
    version="0.0.0rc3",  # Make sure this matches `rai_experiments.__version__`!
    install_requires=INSTALL_REQUIRES,
    python_requires=">=3.7",
    packages=find_packages(where="src", exclude=["tests", "tests.*"]),
    package_dir={"": "src"},
    license="MIT",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
)
